"""
    DotKnot-PW: software for pairwise RNA pseudoknot prediction
    Copyright (C) 2011-2012 Jana Sperschneider	

    This program is free software; you can redistribute it and/or modify  
    it under the terms of the GNU General Public License as published by  
    the Free Software Foundation; either version 2 of the License, or     
    (at your option) any later version.    
"""
import os
import sys
import functions
from operator import itemgetter
import math

def add_pair(interval, score, intervals_pairs, prob):

    if score > 0.0:
        if interval in intervals_pairs:        
            if intervals_pairs[interval] < score:
                intervals_pairs[interval] = score
        else:
            intervals_pairs[interval] = score 
        
    return intervals_pairs
    
                
def collect_intervals(pk_pairs, stem_pairs, conserved_elements):

    all_pk_scores_keys = {}
        
    # Delete duplicate interval matches
    # Key: (pk1_start, pk1_end, pk2_start, pk2_end)
    # Value: score, pk1, pk2
    intervals_pairs = {}
    for (pk1, pk2), (combined_score, ribosum_score, prob, dissim) in pk_pairs.items():
        
        key = pk1[0:2], pk2[0:2]        
        if key in intervals_pairs:
            if intervals_pairs[key] < combined_score:
                intervals_pairs[key] = combined_score
                all_pk_scores_keys[key] = combined_score
        else:
            intervals_pairs[key] = combined_score
            all_pk_scores_keys[key] = combined_score
    
    # Include conserved stem matches 
    for key, (combined_score, ribosum_score, prob, stemlength1, stemlength2, dissim) in stem_pairs.items():        
        intervals_pairs = add_pair(key, combined_score, intervals_pairs, prob)         
        
    # Include conserved secondary structure elements
    if conserved_elements:    
        for (stem1_start, stem1_end, marker1, stem2_start, stem2_end, marker2), (combined_score, prob) in conserved_elements.items():                             
            intervals_pairs = add_pair(((stem1_start, stem1_end), (stem2_start, stem2_end)), combined_score, intervals_pairs, prob)               

    return intervals_pairs, all_pk_scores_keys
    
###################################################################
# Dynamic Programming

""" Implementation as array to avoid expensive dictionary lookup operations
    and reduce memory requirements.
    Each node in first sequence is stored in a sorted knotlist [1,...,n].
    Each node in second sequence is stored in a sorted knotlist [1,...,m].
    Therefore, each node has a defined position in the list.
    Two-dimensional array is used for entry (pi,pj).

     0 p1 p2 ...   pm
    ------------------
    0   0  0  0     0
    p1  0  4  0     2
    p2  0  0  1     1
    ..        ...
    pn  0  0  0     5
    
matrix = [[0, 0, 0, 0], [0, 4, 0, 2], [0, 0, 1, 1], ..., [0, 0, 0, 5]]
Get entry for p1,p2: matrix[1][2]
"""

# Task: initialize the two-dimensional array
def DP_init(interval_pairs):
    
    # Input: structure element lists 1,...,n and 1,...,m
    # We use only endpoints which occur as structure elements
    array = []    
    knotlist1 = []
    knotlist2 = []

    # Store all elements from the first sequence in one list
    # Store all elements from the second sequence in another list
    for element in sorted(interval_pairs):
        key1 = element[0][0], element[0][1]    
        knotlist1.append(key1)
        key2 = element[1][0], element[1][1]    
        knotlist2.append(key2)

    # Delete duplicates    
    knotlist1 = list(set(knotlist1))
    knotlist1.sort()
    knotlist2 = list(set(knotlist2))
    knotlist2.sort()

    knotlist1 = sorted(knotlist1, key=lambda x: x[1])
    knotlist2 = sorted(knotlist2, key=lambda x: x[1])
    
    # Initialize array       
    rows = len(knotlist1)        
    columns = len(knotlist2)
    
    array = [[0.0 for i in range(columns + 1)] for j in range(rows + 1)]                                                    

    # Store list positions in a dictionary for easy lookup
    knotlist1_dic = dict((pk, index + 1) for index, pk in enumerate(knotlist1))
    knotlist2_dic = dict((pk, index + 1) for index, pk in enumerate(knotlist2))    

    return knotlist1, knotlist2, array, knotlist1_dic, knotlist2_dic


# Task: for each interval, store predecessor interval which does not overlap
def predecessor_intervals(knotlist1, knotlist2):
   
    pre_intervals1 = {}
    for position in xrange(len(knotlist1)):
        found = False
        interval = knotlist1[position]
        # Run through all intervals before this one in reverse order
        candidates = knotlist1[:position]        
        candidates.reverse()
        for item in candidates:
            if interval[0] > item[1]:
                pre_intervals1[interval] = item
                found = True
                break
        if not found:
            pre_intervals1[interval] = 0
    
    pre_intervals2 = {}
    for position in xrange(len(knotlist2)):
        found = False
        interval = knotlist2[position]
        # Run through all intervals before this one in reverse order
        candidates = knotlist2[:position]        
        candidates.reverse()
        for item in candidates:
            if interval[0] > item[1]:
                pre_intervals2[interval] = item
                found = True                
                break                
        if not found:
            pre_intervals2[interval] = 0
    
    return pre_intervals1, pre_intervals2


def compatible_intervals(all_nested_stems, loop_start_1, loop_start_2, loop_end_1, loop_end_2, intervals_pairs):
    nested_stems = []
    
    for stem_match in all_nested_stems:
        if stem_match[0][0] >= loop_start_1 and stem_match[0][1] <= loop_end_1:
            if stem_match[1][0] >= loop_start_2 and stem_match[1][1] <= loop_end_2:    
                nested_stems.append(stem_match)

    nested_candidates_dic = {}  
    for match in nested_stems:
        nested_candidates_dic[match] = intervals_pairs[match]   
                
    return nested_candidates_dic
    



def nested_DP(intervals_pairs, conserved_stems, all_pk_scores_keys, stem_dic1, stem_dic2):
    """ Function: DP_nested()

        Purpose:  To pre-compute nested structure elements for all outer hairpin loop
                  intervals via dynamic programming.
                  Given set of structure elements, calculate f(i,j) for each loop 
                  interval [i,j] where i < j using the following recursion:
                  f(i,i) = 0
                               {  f(i,j-1)
                  f(i,j) = max {
                               {  f(i,k-1) + S([k:j]) for all intervals which end at j

                  For each loop interval [i,j], f(i,j) holds the set of nested 
                  intervals which start after or at i and end before or at j.
                  
                  Implementation of dynamic programming matrix as array to avoid
                  expensive dictionary lookup operations and reduce memory requirements.              
                  
        Input:    Set of structure elements and upper loop length limit. 
        
        Return:   Dynamic programming matrix.
    """         
    #------------    
    # Create a dictionary which stores for each right endpoint a
    # list of structure elements which share this right endpoint.         
    endpoints_seq1 = {}
    
    for item in intervals_pairs:
        right_endpoint = item[0][1]              
        if right_endpoint in endpoints_seq1:
            values = endpoints_seq1[right_endpoint]
            values.append(item)
        else:
            values = []
            values.append(item)
        endpoints_seq1[right_endpoint] = values    
    #------------
    """ 
    The two-dimensional array will store all interval pairs which start after point i in the 
    first sequence and end exactly at point j in the first sequence.
    """
    end = max(endpoints_seq1)
    array = [[[] for i in xrange(end + 1)] for j in xrange(end + 1)]                 

    for i in xrange(1, end + 1):
        for j in xrange(i + 1, end + 1):
		
            # Store sets of intervals nested in [i:j] for first sequence in entry array[i][j]            
            previous_value = array[i][j-1]
            added_intervals = []

            # At least one interval ends at endpoint j            
            if j in endpoints_seq1:

                # All intervals which end at j
                list_of_endpoints = endpoints_seq1[j]                               

                for interval_match in list_of_endpoints:                     
                    if interval_match[0][0] >= i:                        
                        added_intervals.append(interval_match)
    
            array[i][j] = added_intervals
    #------------    
    all_stem_scores_nested = {}
     
    significant_stems = [(stem1, stem2) for (stem1, stem2) in conserved_stems if conserved_stems[(stem1, stem2)][2] < 0.01]

    significant_stems_list = sorted(significant_stems, key=lambda x: x[0][1])    

    # Now run through the list of nested stems for sequence 1 and choose those which are compatible with sequence 2
    for (stem1, stem2) in significant_stems_list:
        
        # Watch out because of probabilty filtering
        if (stem1, stem2) in intervals_pairs:                        

            # These are all stems which start after position stem1[0] in the first sequence
            # and end at exactly position stem1[1] in the first sequence
            all_nested_stems = array[stem1[0]][stem1[1]]        
            # Collect all nested stem candidates from previous intervals
            for index in xrange(1, stem1[1]):
                all_nested_stems = all_nested_stems + array[stem1[0]][index]
                    
            # Now check that nested candidate stems are really compatible
            stemlength1 = conserved_stems[(stem1, stem2)][3]
            stemlength2 = conserved_stems[(stem1, stem2)][4]

            # Allow overlap of 1 nt
            overlap = 1
            loop_start_1 = stem1[0] + stemlength1 - overlap
            loop_start_2 = stem2[0] + stemlength2 - overlap       
            loop_end_1 = stem1[1] - stemlength1 + overlap
            loop_end_2 = stem2[1] - stemlength2 + overlap         

            nested_candidates_dic = compatible_intervals(all_nested_stems, loop_start_1, loop_start_2, loop_end_1, loop_end_2, intervals_pairs)           

            if nested_candidates_dic:  

                knotlist1_nested, knotlist2_nested, array_nested, knotlist1_dic_nested, knotlist2_dic_nested = DP_init(nested_candidates_dic)
                pre_intervals1_nested, pre_intervals2_nested = predecessor_intervals(knotlist1_nested, knotlist2_nested)    
                result = DP_bottom_up(knotlist1_nested, knotlist2_nested, array_nested, nested_candidates_dic, pre_intervals1_nested, pre_intervals2_nested, knotlist1_dic_nested, knotlist2_dic_nested)
      
                if result:    
                                
                    conservation_score = sum([value for item, value in result])                                                                              
                                                                                                                   
                    all_stem_scores_nested[stem1, stem2] = conserved_stems[stem1, stem2][0] + conservation_score, result                                                        
    
                    conserved_stems[stem1, stem2] = (conserved_stems[stem1, stem2][0] + conservation_score), conserved_stems[stem1, stem2][1], conserved_stems[stem1, stem2][2], conserved_stems[stem1, stem2][3], conserved_stems[stem1, stem2][4]                                                                                

    return all_stem_scores_nested
    

def traceback(array, knotlist1, knotlist2, predecessor_dic):

    maximum = array[-1][-1]     
    result = []    

    last_interval = None          
    
    # Run through the lists in reverse order
    for i in reversed(knotlist1):
        for a in reversed(knotlist2):            
            candidate = i, a
            if candidate in predecessor_dic:                 
                if round(maximum,10) == round(predecessor_dic[candidate][1],10):
                    if last_interval:
                        if candidate[0][1] < last_interval[0][0] or candidate == last_interval:
                            maximum = round(predecessor_dic[candidate][1],10) - round(predecessor_dic[candidate][0],10)
                            maximum = round(predecessor_dic[candidate][1] - predecessor_dic[candidate][0],10)

                            result.append((candidate, predecessor_dic[candidate][0]))
                            last_interval = candidate
                    else:
                        maximum = round(predecessor_dic[candidate][1],10) - round(predecessor_dic[candidate][0],10)
                        maximum = round(predecessor_dic[candidate][1] - predecessor_dic[candidate][0],10)

                        result.append((candidate, predecessor_dic[candidate][0]))
                        last_interval = candidate
                            
    if maximum != 0.0:                        
        print "Traceback wrong", maximum
        sys.exit()
        
    return result


def DP_bottom_up(knotlist1, knotlist2, array, intervals_pairs, pre_intervals1, pre_intervals2, knotlist1_dic, knotlist2_dic):  
    """ Function: DP_bottom_up()

        Purpose:  Calculate structure element alignment for two given sequences.
                  
        Input:    .
        
        Return:   .
    """
    predecessor_dic = {}
    
    # Run through lists sorted by right endpoint
    for pos1, i in enumerate(knotlist1):
        for pos2, a in enumerate(knotlist2):    

            # Watch out that pos - 1 is never negative in list index            
            score = 0.0

            # Calculate f(i,a)
            # Node i has no match, look up score for f(i-1,a)
            indel1 = array[pos1][pos2 + 1]            

            # Node i has no match, look up score for f(i,a-1)
            indel2 = array[pos1 + 1][pos2] 

            # Node i has a match, include it and look up score for f(pre(i),pre(a))          
            if (i,a) in intervals_pairs:
                match_left1 = pre_intervals1[i]
                match_left2 = pre_intervals2[a]              
                
                if match_left1 and match_left2:                
                    score = intervals_pairs[i, a] + array[knotlist1_dic[match_left1]][knotlist2_dic[match_left2]]	                    
                else:	
                    # Node i and a match, but there is no predecessor		    
                    score = intervals_pairs[i, a] 
                    
            array[pos1 + 1][pos2 + 1] = max(indel1, indel2, score)
                                
            if max(indel1, indel2, score) == score:
                # If node was chosen as best match
                if (i,a) in intervals_pairs:
                    predecessor_dic[i,a] = intervals_pairs[i,a], score, match_left1, match_left2                
                                                
    result = traceback(array, knotlist1, knotlist2, predecessor_dic)    
    
    return result




    
