"""
    DotKnot-PW: software for pairwise RNA pseudoknot prediction
    Copyright (C) 2011-2012 Jana Sperschneider	

    This program is free software; you can redistribute it and/or modify  
    it under the terms of the GNU General Public License as published by  
    the Free Software Foundation; either version 2 of the License, or     
    (at your option) any later version.    
"""

import scores
#---------------------------------------------------------------------------------------------------------             
gev_parameters = {
# stemlength : location_gev, scale_gev, shape_gev               
3 : (6.2656300, 2.1964601, -0.1468699) ,
4 : (7.2147280, 2.4309990, -0.1406325) ,
5 : (8.116595, 2.514450, -0.116540) ,
6 : (7.99878347, 2.51635600, -0.08248844) ,
7 : (8.58702784, 2.40085386, -0.08390027) , 
8 : (9.97894899, 2.77863232, -0.07065527) ,
9 : (10.3113633, 2.54373517, -0.02892308) ,           
10: (10.3113633, 2.54373517, -0.02892308)
}
#--------------------------------------------------------------------------------------------------------- 
def calculate_score_list(stem1, values1, stem2_list, ribosum_bases_first_seq, ribosum_bases_second_seq, ribosum_cache):    
    """ Function: calculate_score()

        Purpose:  Calculate RIBOSUM score.   
                  
        Input:    One target stem with values, one list of query stems and precalculated RIBOSUM bases dictionaries. 
        
        Return:   RIBOSUM score.
    """    
    score_list = []
    
    stemlength1 = values1[0] 
    bps1 = ribosum_bases_first_seq[stem1, stemlength1]  
    key1 = reduce(lambda x,y:10*x+y, bps1)
    
    for (stem2, values2) in stem2_list:
        
        stemlength2 = values2[0]
        bps2 = ribosum_bases_second_seq[stem2, stemlength2]    
        key2 = reduce(lambda x,y:10*x+y, bps2)

        if (key1,key2) in ribosum_cache:
            ribosum_score = ribosum_cache[(key1,key2)]        
        else:
            ribosum_score = scores.ribosum_score_DP(bps1, bps2)        
            ribosum_cache[(key1,key2)] = ribosum_score
                    
        score_list.append(ribosum_score)        
    
    return score_list, ribosum_cache    
#--------------------------------------------------------------------------------------------------------- 
def stem_match_fixed_parameters(first_seq, first_dic_stems, second_seq, second_dic_stems, P_VALUE, conserved_elements):
    """ Function: stem_match_fixed_parameters()

        Purpose:  Two sequences with corresponding stem dictionaries are
                  given. For each stem in first sequence, compute stem
                  similarity score with stems in second sequence.   
                  
        Input:    Two stem dictionaries and two sequences. 
        
        Return:   A dictionary of significant stem similarity scores. A stem dictionary
                  with only the conserved stems for each sequence. A dictionary of
                  RIBOSUM scores for stem pairs for later use.
    """               
    conserved_stems = {}    
    ribosum_cache = {}

    # Ribosum score pre-processing. Store list of base pairs for each stem.
    ribosum_bases_first_seq = scores.ribosum_bases(first_seq, first_dic_stems, None, None)
    ribosum_bases_second_seq = scores.ribosum_bases(second_seq, second_dic_stems, None, None)
    
    # Pre-calculate the RIBOSUM score thresholds with respect to given p-value
    dic_gev_thresholds = scores.thresholds(P_VALUE, gev_parameters)
    
    # -------------------
    for stem1, values1 in first_dic_stems.items():
                                
        stemlength1 = values1[0]

        ribosum_score_list, ribosum_cache = calculate_score_list(stem1, values1, second_dic_stems.items(), ribosum_bases_first_seq, ribosum_bases_second_seq, ribosum_cache)        
        prob_list = [scores.get_prob(stemlength1, ribosum_score_list[index], dic_gev_thresholds, gev_parameters) for index, (stem2, values2) in enumerate(second_dic_stems.items())]
        
        for index, (stem2, values2) in enumerate(second_dic_stems.items()):            
            
            ribosum_score = ribosum_score_list[index]        
            prob = prob_list[index]
            
            if prob < P_VALUE:
                
                stemlength2 = values2[0]
                energy2 = values2[3]
                looplength2 = (stem2[1] - values2[0]) - (stem2[0] + values2[0]) + 1                

                energy1 = values1[3]
                looplength1 = (stem1[1] - stemlength1) - (stem1[0] + stemlength1) + 1
                
                dissim = scores.dissimilarity(stemlength1, stemlength2, looplength1, looplength2)                            
                combined_score = scores.DP_score(ribosum_score, energy1 + energy2, dissim)                                                                                                 
    
                if combined_score > 0.0:
                    conserved_stems[stem1, stem2] = combined_score, ribosum_score, prob, stemlength1, stemlength2, (dissim, energy1, energy2)
                    conserved_elements[(stem1[0], stem1[1], 'hp', stem2[0], stem2[1], 'hp')] = combined_score, ribosum_score
    
    return conserved_stems, conserved_elements


#--------------------------------------------------------------------------------------------------------- 

        
