"""
    DotKnot-PW: software for pairwise RNA pseudoknot prediction
    Copyright (C) 2011-2012 Jana Sperschneider	

    This program is free software; you can redistribute it and/or modify  
    it under the terms of the GNU General Public License as published by  
    the Free Software Foundation; either version 2 of the License, or     
    (at your option) any later version.    
"""

from subprocess import Popen, PIPE
import os
import sys
import user
import pk_construction
import pk_interrupted
import pk_recursive
import mwis
import mwis_2d
import main_function

import find_conserved_stems
import find_conserved_elements
import secondary_elements_modified
import alignment_pk
import cc06, cc09, longPK
import stems
import pk_energy
import DP_array_intervals
import pk_tools

from operator import itemgetter

#--------------------------------------------------------------------------------------------------------- 
P_VALUE = 0.05      # 5% level of significance for secondary structure elements
P_VALUE_PK = 0.01   # 1% level of significance for pseudoknots
#--------------------------------------------------------------------------------------------------------- 

def pairwise_prediction(stem_dic1, bulges_internal1, multiloops1, structures_recursive1, stems_shortened_dic1, pk_alignment_candidates1, i, j, list_of_dictionaries, RNAFOLD_PATH, FILE_ID):
    
    seq1, seq2 = list_of_dictionaries[i], list_of_dictionaries[j]    
    stem_dic2 = main_function.find_stems(seq2, RNAFOLD_PATH, FILE_ID)
    conserved_elements = {}
         
    # Find conserved stems in two sequences
    conserved_stems, conserved_elements = find_conserved_stems.stem_match_fixed_parameters(seq1, stem_dic1, seq2, stem_dic2, P_VALUE, conserved_elements)                                     
    #---------------------------------------------------------------------------------------------------------
    bulges_internal2, multiloops2, structures_recursive2 = compute_secondary_elements(stem_dic2, seq2, RNAFOLD_PATH, FILE_ID)  
    conserved_elements = find_conserved_elements.conserved_bulges(seq1, stem_dic1, bulges_internal1, seq2, stem_dic2, bulges_internal2, P_VALUE, conserved_elements)          
    conserved_elements = find_conserved_elements.conserved_multiloops(seq1, multiloops1, seq2, multiloops2, structures_recursive1, structures_recursive2, conserved_stems, conserved_elements, P_VALUE)
    #---------------------------------------------------------------------------------------------------------
    # Now calculate H-type pseudoknot candidates for each sequence
    stems_shortened_dic2, pk_alignment_candidates2 = dotknot(seq2, stem_dic2, bulges_internal2, multiloops2, RNAFOLD_PATH, FILE_ID) 
    #---------------------------------------------------------------------------------------------------------
    # Look for pseudoknot matches in the first two sequences
    if pk_alignment_candidates1 and pk_alignment_candidates2:
        pk_pairs = alignment_pk.score_pairwise_pks(seq1, bulges_internal1, pk_alignment_candidates1, seq2, bulges_internal2, pk_alignment_candidates2, conserved_elements, P_VALUE_PK)        
    else:
        pk_pairs = {}
    #---------------------------------------------------------------------------------------------------------    
    intervals_pairs, all_pk_scores_keys = DP_array_intervals.collect_intervals(pk_pairs, conserved_stems, conserved_elements)
    #---------------------------------------------------------------------------------------------------------
    all_stem_scores_nested = {}
    if intervals_pairs:
        all_stem_scores_nested = DP_array_intervals.nested_DP(intervals_pairs, conserved_stems, all_pk_scores_keys, stem_dic1, stem_dic2)
    
    # Now update scores for outer stems with nested stems 
    for stem_match, (score, elements) in all_stem_scores_nested.items():        
        if stem_match in intervals_pairs:         
            if score > intervals_pairs[stem_match]:
                intervals_pairs[stem_match] = score   
    #---------------------------------------------------------------------------------------------------------    
    knotlist1, knotlist2, array, knotlist1_dic, knotlist2_dic = DP_array_intervals.DP_init(intervals_pairs) 

    pre_intervals1, pre_intervals2 = DP_array_intervals.predecessor_intervals(knotlist1, knotlist2)

    result = DP_array_intervals.DP_bottom_up(knotlist1, knotlist2, array, intervals_pairs, pre_intervals1, pre_intervals2, knotlist1_dic, knotlist2_dic)

    score_sum = sum([item[1] for item in result])
    
    best_pks_seq1 = []
    secondary_elements_seq1 = []
    
    # Now traceback to check which elements are in nested stems
    best_pks_seq, secondary_elements_seq = print_recursive(result, all_stem_scores_nested, best_pks_seq1, secondary_elements_seq1, pk_pairs, conserved_elements)    
    
    best_pks_seq1 = [pk1 for ((pk1, pk2), values) in best_pks_seq]
    secondary_elements_seq1 = [secondary_element for (secondary_element, values) in secondary_elements_seq]

    best_pks_seq2 = [pk2 for ((pk1, pk2), values) in best_pks_seq]
    secondary_elements_seq2 = [(secondary_element[3], secondary_element[4], secondary_element[5], secondary_element[0], secondary_element[1], secondary_element[2]) for (secondary_element, values) in secondary_elements_seq]
        
    # Print detected pseudoknots
    pseudoknot_list = pk_tools.dot_bracket(seq1, best_pks_seq1, stem_dic1, stems_shortened_dic1, bulges_internal1, multiloops1, pk_alignment_candidates1, {})
    predicted_global_structure = assemble_global_structure(secondary_elements_seq1, pseudoknot_list, seq1, stem_dic1, bulges_internal1, multiloops1, pk_alignment_candidates1) 

    pseudoknot_list_seq2 = pk_tools.dot_bracket(seq2, best_pks_seq2, stem_dic2, stems_shortened_dic2, bulges_internal2, multiloops2, pk_alignment_candidates2, {})
    predicted_global_structure_seq2 = assemble_global_structure(secondary_elements_seq2, pseudoknot_list_seq2, seq2, stem_dic2, bulges_internal2, multiloops2, pk_alignment_candidates2) 

    return pseudoknot_list, result, predicted_global_structure, score_sum, predicted_global_structure_seq2

#--------------------------------------------------------------------------------------------------------- 

def assemble_global_structure(secondary_elements, pseudoknot_list, seq, stem_dic, bulges_internal, multiloops, pseudoknots):

    """ Function: assemble_global_structure()

        Purpose:  Assemble the global structure including pseudoknots.
                  
        Input:    List of predicted pseudoknots and secondary structures. 
        
        Return:   Global structure in dot-bracket notation.
    """
        
    predicted_global_structure = ['.' for i in xrange(len(seq))]

    # Insert secondary structure elements
    if secondary_elements:        
        for element in sorted(secondary_elements):
            # Insert detected hairpin loops
            if element[2] == 'hp':  
                length = stem_dic[element[0], element[1]][0]
                structure = ['.' for i in xrange(element[1] - element[0] + 1)]
                for i in xrange(length):  
                    structure[i] = '('
                    structure[len(structure) - i - 1] = ')'
                
                if predicted_global_structure[element[0] - 1] == '.' and predicted_global_structure[element[1] - 1] == '.':
                    predicted_global_structure[element[0] - 1:element[1]] = structure                
                else:
                    predicted_global_structure[element[0]:element[1] - 1] = structure[1:-1]

            # Insert detected bulge loop, internal loop and multiloop structures
            elif element[2] == 'ib': 
                structure = bulges_internal[element[0], element[1]][1]
                structure = structure.replace(':','')        
                structure = list(structure)
                
                if predicted_global_structure[element[0] - 1] == '.' and predicted_global_structure[element[1] - 1] == '.':                                                                                                
                    predicted_global_structure[element[0] - 1:element[1]] = structure
                else:
                    predicted_global_structure[element[0]:element[1] - 1] = structure[1:-1]

            elif element[2] == 'ml': 
                structure = multiloops[element[0], element[1]][1]
                structure = structure.replace(':','') 
                structure = list(structure)             
                
                if predicted_global_structure[element[0] - 1] == '.' and predicted_global_structure[element[1] - 1] == '.':                                          
                    predicted_global_structure[element[0] - 1:element[1]] = structure
                else:
                    predicted_global_structure[element[0]:element[1] - 1] = structure[1:-1]
                    
    # Insert detected pseudoknots and kissing hairpins

    if pseudoknot_list:
        for pk in pseudoknot_list:  
            
            structure = list(pk[4])
            pk_info = None
            
            # Search for stem lengths etc. in pseudoknot dictionary
            start, end, energy = pk[0], pk[1], pk[2]
            for candidate, values in pseudoknots.items():
                if start == candidate[0] and end == candidate[1]:
                    if round(values[0], 2) == energy:
                        pk_info = candidate
            
            
            # For pseudoknots, cut both stems S1 and S2. Only works for H-type, not kissing hairpins
            if predicted_global_structure[pk[0] - 1] == '.' and predicted_global_structure[pk[1] - 1] == '.':                 
                predicted_global_structure[pk[0] - 1:pk[1]] = structure
            else:
                # There is an overlap 
                shortended_pk = structure
		pk_start = pk[0] - 1
	        pk_end = pk[1]

                if predicted_global_structure[pk[0] - 1] != '.':                               
                    # Cut stem S1 on the left and right side
                    last_bp = pk_info[3] - pk_info[2] 
                    shortended_pk[last_bp] = '.'
                    shortended_pk = shortended_pk[1:]
		    pk_start = pk[0]

                if predicted_global_structure[pk[1] - 1] != '.':                              
                    # Cut stem S2 on the left and right side
                    shortended_pk = shortended_pk[:-1]
                    first_bp = shortended_pk.index('[')
                    shortended_pk[first_bp] = '.'
		    pk_end = pk[1]
               
                predicted_global_structure[pk_start:pk_end] = shortended_pk

    predicted_global_structure = ''.join(predicted_global_structure)        
      
    return predicted_global_structure

#--------------------------------------------------------------------------------------------------------- 
    
def print_recursive(candidate_list, all_stem_scores_nested, best_pks, secondary_elements, all_pk_scores, conserved_elements):
    """ Function: print_recursive()

        Purpose:  Given the DP result, look for nested pseudoknots
                  and secondary structures recursively. Store free 
                  energy information. 
                  
        Input:    Outer structure and set of candidate structures. 
        
        Return:   Nested structures with free energies.  
    """                             

    for (element, score) in candidate_list:

        # Check if element is a pseudoknot pair
        for pk_match, values in all_pk_scores.items():
            if element[0][0] == pk_match[0][0] and element[0][1] == pk_match[0][1]:
                if element[1][0] == pk_match[1][0] and element[1][1] == pk_match[1][1]:
                    match_score = values[0]
                    if score == match_score:
                        best_pks.append((pk_match, score)) 

        if element in all_stem_scores_nested:
            if score == all_stem_scores_nested[element][0]:   # It must be a stem match with updated nested score
                key = element[0][0], element[0][1], 'hp', element[1][0], element[1][1], 'hp'
                secondary_elements.append((key, score))
                
                candidate_list = all_stem_scores_nested[element][1]
                best_pks, secondary_elements = print_recursive(candidate_list, all_stem_scores_nested, best_pks, secondary_elements, all_pk_scores, conserved_elements)           
            else:        
                # Check if element is a secondary structure element
                # Format (stem1[0], stem1[1], 'ib', stem2[0], stem2[1], 'hp')
                for key, values in conserved_elements.items():            
                    if element[0][0] == key[0] and element[0][1] == key[1]:                
                        if element[1][0] == key[3] and element[1][1] == key[4]:                                        
                            if score == values[0]:
                                secondary_elements.append((key, score))   
        else:        
            # Check if element is a secondary structure element
            # Format (stem1[0], stem1[1], 'ib', stem2[0], stem2[1], 'hp')
            for key, values in conserved_elements.items():            
                if element[0][0] == key[0] and element[0][1] == key[1]:                
                    if element[1][0] == key[3] and element[1][1] == key[4]:                                        
                        if score == values[0]:
                            secondary_elements.append((key, score))                           
     
    return best_pks, secondary_elements

#--------------------------------------------------------------------------------------------------------- 
def find_stems(seq, RNAFOLD_PATH, FILE_ID):        
    
    basepairs = stems.RNAfold_dotplot(seq, RNAFOLD_PATH, FILE_ID)            
    stem_dic = stems.find_stems(basepairs)
    stem_dic = stems.evaluation_stems(stem_dic, seq, RNAFOLD_PATH, FILE_ID)    
    CUTOFF_STACK = 0.0 
    CUTOFF_LOOP = 4.0
    stem_dic = stems.filter_stems(stem_dic, CUTOFF_STACK, CUTOFF_LOOP)    

    return stem_dic
#---------------------------------------------------------------------------------------------------------
CUTOFF_IB_ML = 0.0

def compute_secondary_elements(conserved_stems, seq, RNAFOLD_PATH, FILE_ID):

    # Find conserved bulge loops, internal loops and multiloops
    structures_dic = secondary_elements_modified.internal_mwis(conserved_stems)
    bulges_internal, multiloops, structures_recursive = secondary_elements_modified.evaluation_secondary_structures(structures_dic, seq, RNAFOLD_PATH, FILE_ID)
    bulges_internal = secondary_elements_modified.filter_stems(bulges_internal, CUTOFF_IB_ML)
    multiloops = secondary_elements_modified.filter_stems(multiloops, CUTOFF_IB_ML)
    
    return bulges_internal, multiloops, structures_recursive    
#--------------------------------------------------------------------------------------------------------- 

def dotknot(seq, stem_dic, bulges_internal, multiloops, RNAFOLD_PATH, FILE_ID):
    CUTOFF_STACK = 0.0 
    CUTOFF_LOOP = 4.0   
    #---------------------------------------------------------------------------------------------------------
    # Construct H-type core pseudoknots
    stems_shortened, pk_dic = pk_construction.build_pseudoknots(stem_dic)
    # Construct pseudoknots with one interrupted stem
    stems_shortened_ib, pk_dic_ib = pk_construction.build_pseudoknots_ib(bulges_internal, stem_dic)      
    #---------------------------------------------------------------------------------------------------------
    # Re-evaluate stem energies to account for shortended stems in pseudoknots
    stems_shortened.update(stems_shortened_ib)
    if stems_shortened:
        stems_shortened_dic = stems.evaluation_stems(stems_shortened, seq, RNAFOLD_PATH, FILE_ID)
        stems_shortened_dic = stems.update_confidence(stem_dic, stems_shortened_dic)
        stems_shortened_dic = stems.filter_stems(stems_shortened_dic, CUTOFF_STACK, CUTOFF_LOOP)
    else:
        stems_shortened_dic = {}    
    #---------------------------------------------------------------------------------------------------------
    # Construct three different pseudoknot dictionaries
    pk_dic_cc06, pk_dic_cc09, pk_dic_longpk = pk_construction.pk_dic_scan(pk_dic, stem_dic, stems_shortened_dic)
    # Heuristic energy parameters
    INIT = 7.0 
    PENALTY = 0.1
    # Evaluate energies for pseudoknots with loop L2 <= 1nt
    pk_dic_cc06_result = cc06.dic_caochen06(pk_dic_cc06, stem_dic, stems_shortened_dic, seq)
    # Evaluate energies for pseudoknots with loop L2 >= 2nt and L2 <= 6nt
    pk_dic_cc09_result, entropies_dic, entropies_dic_L1, entropies_dic_L3 = cc09.dic_caochen09(pk_dic_cc09, stem_dic)
    # Evaluate energies for pseudoknots with loop L2 >= 7nt
    pk_dic_longpk_result = longPK.dic_longpks(pk_dic_longpk, stem_dic, INIT, PENALTY)
    # Now we form a pseudoknot dictionary out of the three results
    pk_core_dic = {}
    pk_core_dic.update(pk_dic_cc06_result)
    pk_core_dic.update(pk_dic_cc09_result)
    pk_core_dic.update(pk_dic_longpk_result)
    #---------------------------------------------------------------------------------------------------------
    # Create list of all secondary structure elements for dynamic programming MWIS calculation
    # Format is (start, end, weight, free energy, marker)
    elements = [(stem[0], stem[1], values[3], values[3], "hp") for stem, values in stem_dic.items() if values[3] <= 0.0]
    elements = elements + [(stem[0], stem[1], values[2], values[2], "ib") for stem, values in bulges_internal.items()]
    elements = elements + [(stem[0], stem[1], values[2], values[2], "ml") for stem, values in multiloops.items()]    
    elements.sort()
    MAX_LOOP_UPPER = 400
    array_DP = mwis_2d.DP_MWIS(seq, elements, MAX_LOOP_UPPER)
    array_traceback = mwis_2d.traceback_mwis_2d(seq, array_DP, MAX_LOOP_UPPER)
    MAX_LOOP_UPPER = 400
    elements_positive = [(stem[0], stem[1], -1.0/round(values[3],2), values[3], "hp") for stem, values in stem_dic.items() if values[3] > 0.0]
    elements_positive.sort()
    array_DP_positive = mwis_2d.DP_MWIS_positive_weights(seq, elements_positive, MAX_LOOP_UPPER)
    array_traceback_positive = mwis_2d.traceback_mwis_2d_positive(seq, array_DP_positive, MAX_LOOP_UPPER)   
    #---------------------------------------------------------------------------------------------------------
    # Now, look for recursive elements in the pseudoknot loops (MWIS) and add them to dictionary
    pk_core_dic = pk_recursive.recursive_pk(pk_core_dic, seq, array_traceback, array_traceback_positive)
    # Construct three different pseudoknot dictionaries for energy re-evaluation because of the loop entropies
    pk_dic_cc06, pk_dic_cc09, pk_dic_longpk = pk_recursive.pk_dic_scan_recursive(pk_core_dic)  
    pk_dic_cc06_result = cc06.pk_energy_reevaluation_06(pk_dic_cc06)
    pk_dic_cc09_result = cc09.pk_energy_reevaluation_09(pk_dic_cc09, entropies_dic, entropies_dic_L1, entropies_dic_L3)    
    pk_dic_longpk_result = longPK.pk_energy_reevaluation_long(pk_dic_longpk, INIT, PENALTY) 
    pk_recursive_dic = {}    
    pk_recursive_dic.update(pk_dic_cc06_result)
    pk_recursive_dic.update(pk_dic_cc09_result)
    pk_recursive_dic.update(pk_dic_longpk_result)  
    #---------------------------------------------------------------------------------------------------------
    # Evaluate energies for pseudoknots with stems s_ib
    pk_dic_ib = pk_interrupted.evaluate_pk_with_IB(pk_dic_ib, stem_dic, stems_shortened_dic, INIT, PENALTY, seq)
    pk_dic_ib = pk_interrupted.recursive_pk(pk_dic_ib, array_traceback, array_traceback_positive)
    pk_dic_ib = pk_interrupted.re_evaluate_pk_with_IB(pk_dic_ib, INIT, PENALTY)
    pk_dic_ib = pk_recursive.pk_filter(pk_dic_ib)
    # Pseudoknots constructed with stems s_ib are stored in the pseudoknot dictionary
    pk_recursive_dic.update(pk_dic_ib)
    #---------------------------------------------------------------------------------------------------------
    return stems_shortened_dic, pk_recursive_dic
    

