"""
    DotKnot-PW: software for pairwise RNA pseudoknot prediction
    Copyright (C) 2011-2012 Jana Sperschneider	

    This program is free software; you can redistribute it and/or modify  
    it under the terms of the GNU General Public License as published by  
    the Free Software Foundation; either version 2 of the License, or     
    (at your option) any later version.    
"""

from scipy.stats import genextreme
 
#--------------------------------------------------------------------------------------------------------- 
def get_prob(key, score, dic_thresholds, dic_parameters):

    probability = 1.0
    
    if score:
        maximum_key = max(dic_thresholds.keys())
            
        if key in dic_thresholds:                    
            if score >= dic_thresholds[key]:
                location_gev, scale_gev, shape_gev = dic_parameters[key]
                probability = genextreme.sf(score, shape_gev, loc = location_gev, scale = scale_gev)
        else:
            if score >= dic_thresholds[maximum_key]:
                location_gev, scale_gev, shape_gev = dic_parameters[maximum_key]
                probability = genextreme.sf(score, shape_gev, loc = location_gev, scale = scale_gev)            
            
    return probability
#--------------------------------------------------------------------------------------------------------- 
""" Inverse survival function (inverse of sf).    
"""
def thresholds(P_VALUE, dictionary):

    thresholds = {}
    
    for (length, (location_gev, scale_gev, shape_gev)) in dictionary.items():

        thresholds[length] = genextreme.isf(P_VALUE, shape_gev, loc = location_gev, scale = scale_gev)

    return thresholds    
#--------------------------------------------------------------------------------------------------------- 
WEIGHT_CONSERVATION = 0.5
WEIGHT_ENERGY = 1.0 
WEIGHT_DISSIM = 0.5

def DP_score(ribosum_score, free_energy, dissim):
    
    combined_score = WEIGHT_CONSERVATION * ribosum_score - WEIGHT_DISSIM * dissim - WEIGHT_ENERGY * free_energy
        
    return combined_score
#--------------------------------------------------------------------------------------------------------- 
def dissimilarity(stemlength1, stemlength2, looplength1, looplength2):
    """ Function: dissimilarity()

        Purpose:  Dissimilarity calculation using several features.
                  
        Input:    Features like stem lengths, loop lengths...
        
        Return:   Dissimilarity.
    """  
    # Stem length dissimilarity
    dissim_stemlength = difference(stemlength1, stemlength2)
    
    # Loop length dissimilarity    
    dissim_looplength = difference(looplength1, looplength2)        
                                            
    return dissim_stemlength + dissim_looplength
#---------------------------------------------------------------------------------------------------------             
def quickmax(number1, number2):
    """ Function: quickmax()

        Purpose:  Fast calculation of maximum.                
                  
        Input:    Two numbers. 
        
        Return:   Maximum.
    """     
    if number1 >= number2:        
        return number1
    else:
        return number2
#--------------------------------------------------------------------------------------------------------- 
def quickmin(number1, number2):
    """ Function: quickmin()

        Purpose:  Fast calculation of minimum.                
                  
        Input:    Two numbers. 
        
        Return:   Minimum.
    """      
    if number1 < number2:        
        return number1
    else:
        return number2
#--------------------------------------------------------------------------------------------------------- 
def difference(value1, value2):
    """ Function: difference()

        Purpose:  Calculation of plain difference.                
                  
        Input:    Two given values. 
        
        Return:   Plain difference.
    """
    dissim = quickmax(value1, value2) - quickmin(value1, value2)

    return dissim
#--------------------------------------------------------------------------------------------------------- 
"""# RIBOSUM 85-60 matrix
ribosum_bps = {
('AU','AU') :  4.49, ('UA','UA') :  4.97, ('CG','CG') :  5.36, ('GC','GC') :  5.62, ('GU','GU') :  3.47, ('UG','UG') :  3.36,
('AU','UA') :  1.61, ('UA','AU') :  1.61, ('CG','AU') :  1.67, ('GC','AU') :  2.70, ('GU','AU') :  0.59, ('UG','AU') : -0.51,
('AU','CG') :  1.67, ('UA','CG') :  2.75, ('CG','UA') :  2.75, ('GC','UA') :  1.60, ('GU','UA') : -0.57, ('UG','UA') :  1.14,
('AU','GC') :  2.70, ('UA','GC') :  1.60, ('CG','GC') :  2.11, ('GC','CG') :  2.11, ('GU','CG') : -0.27, ('UG','CG') :  1.32,
('AU','GU') :  0.59, ('UA','GU') : -0.57, ('CG','GU') : -0.27, ('GC','GU') :  1.21, ('GU','GC') :  1.21, ('UG','GC') : -0.08,
('AU','UG') : -0.51, ('UA','UG') :  1.14, ('CG','UG') :  1.32, ('GC','UG') : -0.08, ('GU','UG') : -2.09, ('UG','GU') : -2.09 
}"""

# RIBOSUM 85-60 matrix where
# AU == 1
# UA == 2
# CG == 3
# GC == 4
# GU == 5
# UG == 6
bps_ints = {
'AU' :  1, 
'UA' :  2,
'CG' :  3,
'GC' :  4,
'GU' :  5,
'UG' :  6 
}

ribosum_bps = {
(1,1) :  4.49, (2,2) :  4.97, (3,3) :  5.36, (4,4) :  5.62, (5,5) :  3.47, (6,6) :  3.36,
(1,2) :  1.61, (2,1) :  1.61, (3,1) :  1.67, (4,1) :  2.70, (5,1) :  0.59, (6,1) : -0.51,
(1,3) :  1.67, (2,3) :  2.75, (3,2) :  2.75, (4,2) :  1.60, (5,2) : -0.57, (6,2) :  1.14,
(1,4) :  2.70, (2,4) :  1.60, (3,4) :  2.11, (4,3) :  2.11, (5,3) : -0.27, (6,3) :  1.32,
(1,5) :  0.59, (2,5) : -0.57, (3,5) : -0.27, (4,5) :  1.21, (5,4) :  1.21, (6,4) : -0.08,
(1,6) : -0.51, (2,6) :  1.14, (3,6) :  1.32, (4,6) : -0.08, (5,6) : -2.09, (6,5) : -2.09 
}
#--------------------------------------------------------------------------------------------------------- 
    
def ribosum_bases(seq, dic_stems, dic_stems_short, dic_stems_bulges):
    """ Function: ribosum_bases()

        Purpose:  For a given sequence and secondary structure dictionaries,
                  store stem with corresponding list of base pairs for faster
                  RIBOSUM calculation later. Convert this to integers to save space.
                  Example for a stem:  ((1, 20), 3) ['GU', 'CG', 'GC']
                  
        Input:    Sequence and secondary structure dictionaries.                   
        
        Return:   Dictionary which stores element with list of base pairs. 
    """     
    combined_dic_stems = {}
    
    if dic_stems:
        combined_dic_stems.update(dic_stems)    
    if dic_stems_short:
        combined_dic_stems.update(dic_stems_short)

    ribosum_bases = {}
    list(seq)

    """ Store base pairs for regular stems """
    for stem, values in combined_dic_stems.items():
        stemlength = values[0]
        bases_int = []        
        for i in xrange(stemlength):
            basepair = seq[stem[0] + i - 1] + seq[stem[1] - i - 1]                       
            bases_int.append(bps_ints[basepair])
        # Store all the information, watch out for shortened stems
        ribosum_bases[stem[0:2], stemlength] = bases_int

    """ Merge base pairs in interrupted stems by deleting bulges and internal loops """
    if dic_stems_bulges:
        for stem, values in dic_stems_bulges.items():
            bases_int = []
            
            stem_structure = dic_stems_bulges[stem][1]
            stem_structure = stem_structure.replace(':','')
            stem_sequence = seq[stem[0]-1:stem[1]] 
    
            # Merge interrupted stem into consecutive stem
            # Left part of stem
            left_stem_structure = stem_structure[:stem_structure.rfind('(')+1]
            left_stem_sequence = stem_sequence[:stem_structure.rfind('(')+1]    
            # Now remove all unpaired bases in stem
            left_stem_sequence_merged, left_stem_structure_merged = remove_unpaired(left_stem_sequence, left_stem_structure)
            
            # Right part of stem
            right_stem_structure = stem_structure[stem_structure.find(')'):]
            right_stem_sequence = stem_sequence[stem_structure.find(')'):]    
            # Now remove all unpaired bases in stem  
            right_stem_sequence_merged, right_stem_structure_merged = remove_unpaired(right_stem_sequence, right_stem_structure)

            stemlength = left_stem_structure_merged.rfind('(') + 1

            for i in xrange(stemlength):
                basepair = left_stem_sequence_merged[i] + right_stem_sequence_merged[-1 - i]
                bases_int.append(bps_ints[basepair])

            # Store all the information, watch out for shortened stems
            ribosum_bases[stem[0:2], stemlength] = bases_int            
    
    return ribosum_bases
#--------------------------------------------------------------------------------------------------------- 
def ribosum_score_DP(bps1, bps2):
    """ Function: ribosum_score()

        Purpose:  Calculcate best RIBOSUM score for two given stems using
                  a local stem alignment strategy without gaps. 
                  
        Input:    Two stems given as corresponding list of base pairs.
                  Example input for a stem: ['GU', 'CG', 'GC']
        
        Return:   RIBOSUM score. 
    """      
    stemlength1, stemlength2 = len(bps1), len(bps2)
    
    """ Init array
        R(0,i) = 0
        R(j,0) = 0
    """

    array = [[0.0 for i in range(stemlength2 + 1)] for j in range(stemlength1 + 1)]                                                        

    max_score = 0.0
    
    for i in xrange(stemlength1 + 1):
        for j in xrange(stemlength2 + 1):

            pos1 = i - 1
            pos2 = j - 1
            
            if i == 0 or j == 0:
                pass

            else:
                array[i][j] = max(0.0, array[i-1][j-1] + ribosum_bps[(bps1[pos1], bps2[pos2])])

                if array[i][j] > max_score:
                    max_score = array[i][j]                      

    return max_score
#--------------------------------------------------------------------------------------------------------- 

#--------------------------------------------------------------------------------------------------------- 
def remove_unpaired(stem_sequence, stem_structure):
    """ Function: remove_unpaired()

        Purpose:  For an interrupted stem given by its
                  dot-bracket notation, remove all 
                  unpaired bases in structure and sequence. 
                  
        Input:    Interrupted stem sequence and structure.
        
        Return:   Merged stem sequence and structure. 
    """  
    stem_structure_merged = list(stem_structure)
    stem_sequence_merged = list(stem_sequence)
    
    for position in xrange(0,len(stem_structure),1):        
        if stem_structure[position] == '.':
            stem_sequence_merged[position] = '.'
    stem_structure_merged = "".join(stem_structure_merged)
    stem_structure_merged = stem_structure_merged.replace('.','')
    stem_sequence_merged = "".join(stem_sequence_merged)            
    stem_sequence_merged = stem_sequence_merged.replace('.','')  
    
    return stem_sequence_merged, stem_structure_merged
#--------------------------------------------------------------------------------------------------------- 
def merge_interrupted_stem(seq, stem, stem_structure):    
    """ Function: merge_interrupted_stem()

        Purpose:  For a given interrupted stem, output merged
                  stem with all unpaired positions deleted and 
                  base pairs preserved.
                  
        Input:    Sequence and interrupted stem
        
        Return:   Merged stem. 
    """    
    stem_structure = stem_structure.replace(':','')
    stem_sequence = seq[stem[0]-1:stem[1]]

    # Merge interrupted stem into consecutive stem
    # Left part of stem
    left_stem_structure = stem_structure[:stem_structure.rfind('(')+1]
    left_stem_sequence = stem_sequence[:stem_structure.rfind('(')+1]    
    # Now remove all unpaired bases in stem
    left_stem_sequence_merged, left_stem_structure_merged = remove_unpaired(left_stem_sequence, left_stem_structure)
    
    # Right part of stem
    right_stem_structure = stem_structure[stem_structure.find(')'):]
    right_stem_sequence = stem_sequence[stem_structure.find(')'):]    
    # Now remove all unpaired bases in stem  
    right_stem_sequence_merged, right_stem_structure_merged = remove_unpaired(right_stem_sequence, right_stem_structure)

    # Assemble whole stem
    stem_sequence_merged = left_stem_sequence_merged + stem_sequence[stem_structure.rfind('(')+1:stem_structure.find(')')] + right_stem_sequence_merged    
    stem_structure_merged = left_stem_structure_merged + stem_structure[stem_structure.rfind('(')+1:stem_structure.find(')')] + right_stem_structure_merged
    
    return stem_sequence_merged, stem_structure_merged
#--------------------------------------------------------------------------------------------------------- 
_cache = {}

def inner_score_DP(inner_stems1, inner_stems2, conserved_elements):
    """ Function: inner_score_DP()

        Purpose:  Calculcate best RIBOSUM score for two given stems using
                  a local stem alignment strategy, including gaps. 
                  
        Input:    Two stems given as corresponding list of base pairs.
                  Example input for a stem: ['GU', 'CG', 'GC']
        
        Return:   Conservation score. 
    """      
    length1, length2 = len(inner_stems1), len(inner_stems2)  

    inner_stems1 = sorted(inner_stems1)
    inner_stems2 = sorted(inner_stems2)

    left_end1 = inner_stems1[0][0]
    left_end2 = inner_stems2[0][0]    
    right_end1 = inner_stems1[-1][1]
    right_end2 = inner_stems2[-1][1]

    cache_key = (left_end1, right_end1, left_end2, right_end2)
    
    if cache_key in _cache:
        max_score = _cache[cache_key]

    else:
    
        """ Init array
            R(0,i) = 0
            R(j,0) = 0
        """                
        array = [[0.0 for i in range(length2 + 1)] for j in range(length1 + 1)]                                                        

        max_score = 0.0
        
        for i in xrange(length1 + 1):
            for j in xrange(length2 + 1):

                pos1 = i - 1
                pos2 = j - 1

                element1 = inner_stems1[pos1]        
                element2 = inner_stems2[pos2]
                key = element1[0], element1[1], element1[-1], element2[0], element2[1], element2[-1]
                
                if i == 0 or j == 0:
                    pass

                else:
                    local_score = 0.0
                    
                    if key in conserved_elements:
                        local_score = conserved_elements[key][1]                        
                                        
                    array[i][j] = max(0.0, array[i-1][j-1] + local_score, array[i-1][j], array[i][j-1])

                    if array[i][j] > max_score:
                        max_score = array[i][j]
        
        _cache[cache_key] = max_score    

    return max_score
#--------------------------------------------------------------------------------------------------------- 

